function SafeTocFile(name)
	local safepath = name:gsub("\\", "/")
	safepath = safepath:gsub("%.%./", "")
	safepath = safepath:gsub("^%a:/", "")
	safepath = safepath:gsub("^/", "")
	return safepath
end

function ParseTocLine(line, data)
    if line:find("##") then
        for attr, val in line:gfind("(%S+):[%s]*(.+)") do
            --print(attr .. "=" .. val)
            if attr == "Interface" then
				data.interface = tonumber(val)
			elseif attr == "Version" then
				data.version = {}
				for ve in val:gfind("[%d]+") do
					table.insert(data.version, tonumber(ve))
				end
            elseif attr == "SavedVariables" then
				data.savedvars = {}
				for varname in val:gfind("[%w_]+") do
					table.insert(data.savedvars, varname)
				end
			elseif attr == "Dependencies" then
				data.depends = {}
				-- ignore the versions for now
				for dep,vers in val:gfind("([%a_]+)") do
					table.insert(data.depends, dep)
				end
			else
				data[attr:lower()] = val
			end
        end
    elseif line:find("%.lua$") then
		if data.luafiles == nil then
			data.luafiles = {}
		end
		table.insert(data.luafiles, SafeTocFile(line))
	elseif line:find("%.cfg$") then
		if data.cfgfiles == nil then
			data.cfgfiles = {}
		end
		table.insert(data.cfgfiles, SafeTocFile(line))
	end
end

function ParseTocFile(filename)
    local f = io.open(filename)
    local data = {}
	data.luafiles = {}
	data.cfgfiles = {}
    for line in f:lines() do
		ParseTocLine(line, data)
    end
    f:close()
	return data
end

function PrintTocData(data)
	print("===Printing TOC file===")
	for k,v in pairs(data) do
		print(string.format("%s=%s", tostring(k), tostring(v)))
		if type(v) == "table" then
			for key,val in pairs(v) do
				print(string.format("\t%s=%s", tostring(key), tostring(val)))
			end
		end
	end
end

function WriteParseTocLib()
	infile = io.open('parsetoc.lua', 'r')
	outfile = io.open('parsetoc.txt', 'w')
	for line in infile:lines() do
		local formattedsz = string.format("%q \"\\n\"\n", line)
		outfile:write(formattedsz)
		print(formattedsz)
	end
	infile:close()
	outfile:close()
end

