"-- Lua class bind module - implements most of logic code for classbind." "\n"
"" "\n"
"-- params: " "\n"
"--	state=table to store data, stop iteration, etc." "\n"
"--	mt=metatable containing a parents field representing array of leaf nodes" "\n"
"--	fn=function to perform on metatable (given state and metatable params)" "\n"
"-- returns: the tree path from start to end when iteration was stopped, or nothing if the iteration completes" "\n"
"function each_parent(state, mt, fn)" "\n"
"	local level = 1" "\n"
"	local list = { }" "\n"
"	local x,y" "\n"
"	local stop = false" "\n"
"	local function p_visit(st, mt, fn)" "\n"
"		if mt.incomplete then " "\n"
"			error(\"Node \" .. mt.classname .. \" is undefined\")" "\n"
"		end" "\n"
"		list [ level ] = mt" "\n"
"		-- print(\"luacb: visiting\", mt.classname, level)" "\n"
"		stop = fn(st, mt)" "\n"
"		if stop then" "\n"
"			return mt, list" "\n"
"		end" "\n"
"		level = level + 1" "\n"
"		for i, each in ipairs(mt.parents) do" "\n"
"			-- perhaps we need to return?" "\n"
"			x,y = p_visit(st, each, fn)" "\n"
"			if x then return x,y end" "\n"
"		end" "\n"
"		list[ level ] = nil" "\n"
"		level = level - 1" "\n"
"	end" "\n"
"	return p_visit(state, mt, fn)" "\n"
"end" "\n"
"" "\n"
"function each_child(state, mt, fn)" "\n"
"	local level = 1" "\n"
"	local list = { }" "\n"
"	local x,y" "\n"
"	local stop = false" "\n"
"	local function p_visit(st, mt, fn)" "\n"
"		if mt.incomplete then " "\n"
"			error(\"Node \" .. mt.classname .. \" is undefined\")" "\n"
"		end" "\n"
"		list [ level ] = mt" "\n"
"		-- print(\"luacb: visiting\", mt.classname, level)" "\n"
"		stop = fn(st, mt)" "\n"
"		if stop then" "\n"
"			return mt, list" "\n"
"		end" "\n"
"		level = level + 1" "\n"
"		if mt.children == nil then" "\n"
"			return" "\n"
"		end" "\n"
"		for i, each in ipairs(mt.children) do" "\n"
"			-- perhaps we need to return?" "\n"
"			x,y = p_visit(st, each, fn)" "\n"
"			if x then return x,y end" "\n"
"		end" "\n"
"		list[ level ] = nil" "\n"
"		level = level - 1" "\n"
"	end" "\n"
"	return p_visit(state, mt, fn)" "\n"
"end" "\n"
"" "\n"
"-- returns a light user data representing the actual C++ object. uses" "\n"
"-- class hierarchy iteration to perform downcast (if needed). On failure," "\n"
"-- returns the following error codes," "\n"
"-- -1 : not a class ref." "\n"
"-- -2 : parent class not found (downcast fail)" "\n"
"function checkobject(obj, clsname)" "\n"
"	if not isclassref(obj) then " "\n"
"		return -1" "\n"
"	end" "\n"
"	-- is this an exact match?" "\n"
"	if getmt(obj).classname == clsname then " "\n"
"		return getobjptr(obj)" "\n"
"	end" "\n"
"	-- can it be downcasted to a parent class?" "\n"
"	local iter_state = { }" "\n"
"	local lastmt, list = each_parent(iter_state, getmt(obj), function(st, mt)" "\n"
"		if clsname == mt.classname then " "\n"
"			return true" "\n"
"		end" "\n"
"	end )" "\n"
"	if list ~= nil then " "\n"
"		-- found it, so perform the downcast" "\n"
"		local dc = 0" "\n"
"		local i = 1" "\n"
"		while i <= #list - 1 do" "\n"
"			dc = dc + list[i].downcast[ list[i+1].classname ]" "\n"
"			i = i + 1" "\n"
"		end" "\n"
"		-- TODO: cache the result in a downcast table if this operation proves to be" "\n"
"		-- too time consuming." "\n"
"		return topointer( tointeger(getobjptr(obj)) + dc )" "\n"
"	end" "\n"
"	return -2" "\n"
"end" "\n"
"" "\n"
"function typeof(obj)" "\n"
"	if not isclassref(obj) then" "\n"
"		return nil" "\n"
"	end" "\n"
"	return getmt(obj).classname" "\n"
"end" "\n"
"" "\n"
"function registerclass(name, methods, metamethods)" "\n"
"	if getregistry(name) ~= nil and not getregistry(name).incomplete then" "\n"
"		error(\"Attempt to overwrite existing class \" .. name)" "\n"
"	end" "\n"
"	-- print(\"luacb: Making new lib\")" "\n"
"	local libtbl = newlib(name, methods)" "\n"
"	setmt(libtbl, {" "\n"
"		__call = function(tbl, ...)" "\n"
"			return tbl.new(...)" "\n"
"		end" "\n"
"	})" "\n"
"	-- print(\"luacb: Making new mt\")" "\n"
"	newmt(name, metamethods)" "\n"
"	local mt = getregistry(name)" "\n"
"	" "\n"
"	-- initialize common metatable lists" "\n"
"	mt.magic = getclassrefsig()" "\n"
"	mt.classname = name" "\n"
"	mt.parents = mt.parents or {}" "\n"
"	mt.cppobjects = {}" "\n"
"	setmt(mt.cppobjects, { __mode = \"v\" })" "\n"
"	mt.luaobjects = {}" "\n"
"	setmt(mt.luaobjects, { __mode = \"v\" })" "\n"
"	mt.downcast = mt.downcast or {}" "\n"
"	mt.fields = {}" "\n"
"	mt.get = {}" "\n"
"	mt.set = {}" "\n"
"	if mt.incomplete then" "\n"
"		print(\"Class \" .. name .. \" is no longer incomplete\")" "\n"
"	end" "\n"
"	mt.incomplete = nil" "\n"
"	" "\n"
"	-- give a reasonable \"ClassName 0xpointer\" tostring mm if none is provided" "\n"
"	mt.__tostring = mt.__tostring or function(obj)" "\n"
"		return sfmt(\"%s 0x%x\", getobjclsname(obj), tointeger(getobjptr(obj)))" "\n"
"	end" "\n"
"" "\n"
"	-- initialize new and delete methods (if any)" "\n"
"	" "\n"
"	-- a new w/o delete, and vice versa is invalid" "\n"
"	if mt.__new ~= nil and mt.__delete == nil then " "\n"
"		error(\"Class \" .. name .. \" defined new with no delete\")" "\n"
"	end" "\n"
"	" "\n"
"	if mt.__new == nil and mt.__delete ~= nil then " "\n"
"		error(\"Class \" .. name .. \" defined delete with no new\")" "\n"
"	end" "\n"
"	" "\n"
"	if mt.__new ~= nil then" "\n"
"		libtbl.new = function(...)" "\n"
"			local ptr = mt.__new(...)" "\n"
"			if tointeger(ptr) == 0 then " "\n"
"				error(\"New pointer is NULL\")" "\n"
"			end" "\n"
"			local val = newclassref(ptr)" "\n"
"			setmt(val, mt)" "\n"
"			setobjfenv(val, {}) -- all objects have an empty f-env accessible via .object" "\n"
"			mt.luaobjects[ptr] = val" "\n"
"			setneedsdelete(val, true)" "\n"
"			if mt.__ctor ~= nil then" "\n"
"				mt.__ctor(val)" "\n"
"			end" "\n"
"			print(sfmt(\"luacb: new %s 0x%x\", name, tointeger(ptr)))" "\n"
"			return val" "\n"
"		end" "\n"
"	end" "\n"
"	" "\n"
"	mt.__gc = function(obj)" "\n"
"		print(\"luacb: __gc \" .. sfmt(\"%s 0x%x\", name, tointeger(getobjptr(obj)) ))" "\n"
"		-- no need to collect cppobjects, only lua objects" "\n"
"		local ptr = getobjptr(obj)" "\n"
"		if needsdelete( obj ) then" "\n"
"			-- if there is no defined __delete for this class, it is a memory leak" "\n"
"			if mt.__delete == nil then " "\n"
"				error(\"Memory leak in class \" .. getobjclsname(obj))" "\n"
"			end" "\n"
"			if mt.__dtor ~= nil then" "\n"
"				mt.__dtor(obj)" "\n"
"			end" "\n"
"			print(sfmt(\"luacb: deleting %s 0x%x\", name, tointeger(ptr)))" "\n"
"			mt.__delete( ptr )" "\n"
"		end" "\n"
"	end" "\n"
"" "\n"
"	-- initialize the __index and __newindex metamethods" "\n"
"	mt.__index = function(obj, key)" "\n"
"		-- look in the f-env first always:" "\n"
"		if getobjfenv(obj)[key] ~= nil then " "\n"
"			return getobjfenv(obj)[key]" "\n"
"		end" "\n"
"		-- else traverse class hierarchy, looking for class methods, fields, and properties" "\n"
"		local iter_st = {}" "\n"
"		each_parent(iter_st, getmt(obj), function(st, mt)" "\n"
"			-- is it a class method?" "\n"
"			if getlib(mt.classname)[key] ~= nil then" "\n"
"				st.value = getlib(mt.classname)[key]" "\n"
"				return true" "\n"
"			end" "\n"
"			if mt.get[key] then " "\n"
"				st.value = mt.get[key](obj, key)" "\n"
"				return true" "\n"
"			end" "\n"
"			if mt.fields[key] then " "\n"
"				st.value = getcppfield(obj, mt.fields[key])" "\n"
"				return true" "\n"
"			end" "\n"
"		end)" "\n"
"		if iter_st.value == nil then" "\n"
"			-- try object properties" "\n"
"			if key == \"classname\" then" "\n"
"				iter_st.value = mt.classname" "\n"
"			elseif key == \"is\" then" "\n"
"				iter_st.value = function(obj1, obj2)" "\n"
"					return (getobjptr(obj1) == getobjptr(obj2))" "\n"
"				end" "\n"
"			end" "\n"
"		end" "\n"
"		return iter_st.value" "\n"
"	end" "\n"
"	" "\n"
"	mt.__newindex = function(obj, key, val)" "\n"
"		local iter_st = {}" "\n"
"		each_parent(iter_st, getmt(obj), function(st, mt)" "\n"
"			-- is it a set property?" "\n"
"			if mt.set[key] then " "\n"
"				st.success = pcall( mt.set[key], obj, val)" "\n"
"				if st.success then " "\n"
"					return true" "\n"
"				end" "\n"
"			end" "\n"
"			if mt.fields[key] then " "\n"
"				st.success = pcall( setcppfield, obj, mt.fields[key], val)" "\n"
"				if st.success then " "\n"
"					return true" "\n"
"				end" "\n"
"			end" "\n"
"		end)" "\n"
"		if not iter_st.success then " "\n"
"			getobjfenv(obj)[key] = val" "\n"
"		end" "\n"
"	end" "\n"
"" "\n"
"	local fakemt = {}" "\n"
"	local fakeaccess = function()" "\n"
"		error(\"Attempt to access metatable of class \" .. name)" "\n"
"	end" "\n"
"	setmt(fakemt, { __index=fakeaccess, __newindex=fakeaccess })" "\n"
"	mt.__metatable = fakemt" "\n"
"end" "\n"
"" "\n"
"function addparent(cls, parentname, downbytes)" "\n"
"	-- allow deferred parent classes" "\n"
"	if getregistry(cls) == nil then " "\n"
"		printf(\"Warning: incomplete child class \" .. cls)" "\n"
"		local unknowncls = newmt(cls)" "\n"
"		unknowncls.name = cls" "\n"
"		unknowncls.downcast = {}" "\n"
"		unknowncls.parents = {}" "\n"
"		unknowncls.incomplete = true" "\n"
"	end" "\n"
"	" "\n"
"	local mt = getregistry(parentname)" "\n"
"	if mt == nil then" "\n"
"		mt = newmt(parentname)" "\n"
"		print(\"Warning, adding incomplete parent \" .. parentname .. \" to class \" .. cls)" "\n"
"		mt.incomplete = true" "\n"
"	end" "\n"
"	local clsparents = getregistry(cls).parents" "\n"
"	local clsdowncast = getregistry(cls).downcast" "\n"
"	print(sfmt(\"luacb: %s to %s %d bytes\", cls, parentname, downbytes))" "\n"
"	clsdowncast[parentname] = downbytes" "\n"
"	clsparents[#clsparents + 1] = mt" "\n"
"	mt.children = mt.children or {}" "\n"
"	mt.children[#mt.children + 1] = getregistry(cls)" "\n"
"end" "\n"
"" "\n"
"function findobjectvalue(cls, ptr)" "\n"
"	local mt = getregistry(cls)" "\n"
"" "\n"
"	local state = {}" "\n"
"	each_parent(state,mt, function(st, mt)" "\n"
"		if mt.luaobjects[ptr] then" "\n"
"			st.value = mt.luaobjects[ptr]" "\n"
"			return true" "\n"
"		elseif mt.cppobjects[ptr] then " "\n"
"			st.value = mt.cppobjects[ptr]" "\n"
"			return true" "\n"
"		end" "\n"
"	end)" "\n"
"	if state.value == nil and mt.children ~= nil then" "\n"
"		each_child(state, mt, function(st, mt)" "\n"
"			if mt.luaobjects[ptr] then" "\n"
"				st.value = mt.luaobjects[ptr]" "\n"
"				return true" "\n"
"			elseif mt.cppobjects[ptr] then " "\n"
"				st.value = mt.cppobjects[ptr]" "\n"
"				return true" "\n"
"			end			" "\n"
"		end )" "\n"
"	end	" "\n"
"	return state.value" "\n"
"end" "\n"
"" "\n"
"function newcppobject(cls, ptr)" "\n"
"	local value = newclassref(ptr)" "\n"
"	local mt = getregistry(cls)" "\n"
"	if not mt then " "\n"
"		return error(\"Class \" .. cls .. \" is not registered\")" "\n"
"	end" "\n"
"	setmt(value, mt)" "\n"
"	mt.cppobjects[ ptr ] = value" "\n"
"	setneedsdelete(value, false)" "\n"
"	return value" "\n"
"end" "\n"
"" "\n"
"function removeobject(cls, ptr)" "\n"
"	local mt = getregistry(cls)" "\n"
"	mt.cppobjects[ptr] = nil" "\n"
"end" "\n"
"" "\n"
"function bindgetproperty(cls, propname, propfunc)" "\n"
"	getregistry(cls).get[propname] = propfunc" "\n"
"end" "\n"
"" "\n"
"function bindsetproperty(cls, propname, propfunc)" "\n"
"	getregistry(cls).set[propname] = propfunc" "\n"
"end" "\n"
