// the initialization of some of these may be scattered, unfortunately
//

#include "engine.h"
#include "luaclassbind.h"
#include "luaclassbindutil.h"

void push_vec(lua_State* L, const vec& v)
{
	lua_newtable(L);
	lua_pushnumber(L, v.x);
	lua_setfield(L, -2, "x");
	lua_pushnumber(L, v.y);
	lua_setfield(L, -2, "y");
	lua_pushnumber(L, v.z);
	lua_setfield(L, -2, "z");
}

#ifndef STANDALONE
typedef physent PhysEnt;

ClassProperty(get, PhysEnt, o)
{
	PhysEnt* self = CLASSBIND_CHECKOBJECT(L, 1, PhysEnt);
	push_vec(L, self->o);
	return 1;
}

ClassProperty(get, PhysEnt, vel)
{
	PhysEnt* self = CLASSBIND_CHECKOBJECT(L, 1, PhysEnt);
	push_vec(L, self->vel);
	return 1;
}

ClassProperty(get, PhysEnt, falling)
{
	PhysEnt* self = CLASSBIND_CHECKOBJECT(L, 1, PhysEnt);
	push_vec(L, self->falling);
	return 1;
}

int luaopen_class_PhysEnt(lua_State* L)
{
	classbind::registerclass(L, "PhysEnt", NULL,NULL);
	classbind::aliasclass(L, "PhysEnt", "physent");
	BindClassProperty(get, PhysEnt, o);
	BindClassProperty(get, PhysEnt, vel);
	BindClassProperty(get, PhysEnt, falling);
    // vec deltapos, newpos;
	CLASSBIND_BINDCPPFIELD(L, PhysEnt, yaw, float, true);
	CLASSBIND_BINDCPPFIELD(L, PhysEnt, pitch, float, true);
	CLASSBIND_BINDCPPFIELD(L, PhysEnt, roll, float, true);
	CLASSBIND_BINDCPPFIELD(L, PhysEnt, maxspeed, float, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, timeinair, int32, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, radius, float, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, eyeheight, float, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, aboveeye, float, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, xradius, float, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, yradius, float, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, zmargin, float, true);
    // vec floor;
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, inwater, int32, true);
    // bool jumping;
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, move, int8, true);
    CLASSBIND_BINDCPPFIELD(L, PhysEnt, strafe, int8, true);
	/*
    uchar physstate;
    uchar state, editstate;
    uchar type;
    uchar collidetype;
    */
    // bool blocked, moving;
    // physent *onplayer;
//    CLASSBIND_BINDCPPFIELD(L, PhysEnt, lastmove, int32, true);
    //CLASSBIND_BINDCPPFIELD(L, PhysEnt, lastmoveattempt, int32, true);
	return 0;
}

typedef dynent DynEnt;

int luaopen_class_DynEnt(lua_State* L)
{
	classbind::registerclass(L, "DynEnt", NULL, NULL);
	classbind::aliasclass(L, "DynEnt", "dynent");
	CLASSBIND_ADDPARENT(L, DynEnt, PhysEnt);
	// bool k_left, k_right, k_up, k_down;
	// CLASSBIND_BINDCPPFIELD(L, DynEnt, targetyaw, float, true);
	// CLASSBIND_BINDCPPFIELD(L, DynEnt, rotspeed, float, true);
    // entitylight light;
    // animinterpinfo animinterp[MAXANIMPARTS];
    // ragdolldata *ragdoll;
    // occludequery *query;
    CLASSBIND_BINDCPPFIELD(L, DynEnt, occluded, int32, true);
    CLASSBIND_BINDCPPFIELD(L, DynEnt, lastrendered, int32, true);
	return 0;
}

ADDLUAUSERCLASS(PhysEnt);
ADDLUAUSERCLASS(DynEnt);

#endif

/* GetPath(path : string) : string
 |---------------------------------
 | Convert a platform-specific path specifier into a platform-neutral path.
 | (Changes '/' or '\\' into '/'.)
*/
int lua_GetPath(lua_State* L)
{
	const char* arg = luaL_checkstring(L, 1);
	char* result = path(arg, true);
	if(result != 0)
	{
		lua_pushstring(L, result);
		// DELETEA(result); // honestly don't know what to do with the result string
	}
	else
	{
		lua_pushnil(L);
	}
	return 1;
}
ADDLUAUSERAPI(GetPath); // works on any string not just valid file paths

/* CleanString(path) : string
 |----------------------------
 | Return a copy of $path with unsafe path specifiers removed
*/
int lua_CleanString(lua_State* L)
{
	luaL_checkstring(L, 1);
	const char* filters[] = { "\\", "%.%./", "^%a:/", "^/" 	};
	const char* subst[] = {	  "/", 	"", 	 "", 	  "" 	};

	lua_getfield(L, 1, "gsub"); // string's metatable should look this up

	for(uint eachFilter=0; eachFilter < sizeof(filters) / sizeof(char*); ++eachFilter)
	{
		lua_pushvalue(L, -1);
		lua_pushvalue(L, 1);
		lua_pushstring(L, filters[eachFilter]);
		lua_pushstring(L, subst[eachFilter]);
		lua_call(L, 3, 1);
		lua_replace(L, 1);
	}
	lua_pop(L, 1); // remove gsub from stack
	lua_pushvalue(L, 1); // push modified string arg and return it
	return 1;
}
ADDLUAUSERAPI(CleanString);

/* IsDirectory(path : string) : boolean
 |--------------------------------------
 | Returns true if the path is a directory
 |
*/
int lua_IsDirectory(lua_State* L)
{
	vector<char*> files;

	luaL_checkstring(L, 1);
	lua_pushcfunction(L, lua_CleanString);
	lua_pushvalue(L, 1);
	lua_call(L, 1, 1);
	lua_replace(L, 1);
	lua_pushboolean(L, listdir(lua_tostring(L, 1), NULL, files)? 1 : 0);
	return 1;
}
ADDLUAUSERAPI(IsDirectory);

/* GetDirectoryList(dir : string, [ext : string]) : table
 |--------------------------------------------------------
 | Return an array of the contents of directory $dir. If $ext is specified,
 | returns only an array of file names in the directory ending with $ext
 | with the '.' and the extension removed. Any part of the path that
 | attempts to access folders outside of game root folder are string
 | substituted.
*/
int lua_GetDirectoryList(lua_State* L)
{
	luaL_checkstring(L, 1);
	const char* ext = luaL_optstring(L, 2, NULL);

	vector<char*> files;

	// clean the string parameter
	//
	lua_pushcfunction(L, lua_CleanString);
	lua_pushvalue(L, 1);
	lua_call(L, 1, 1);
	lua_replace(L, 1);

	bool ok = listdir(lua_tostring(L, 1),ext,files);

	if(!ok)
	{
		lua_pushnil(L); // no files
	}
	else // result will be empty table if valid empty directory
	{
		lua_newtable(L);
		loopv(files)
		{
			lua_pushinteger(L, i+1); // lua array
			lua_pushstring(L, files[i]);
			lua_settable(L, -3);
		}
	}
	return 1;
}
ADDLUAUSERAPI(GetDirectoryList);

/* DirectoryList()
 |-----------------
 | An iterator version of GetDirectoryList
*/
static int lua_DirectoryList_iter(lua_State* L)
{
	// upvalue1 is a vector<char*>**
	// upvalue2 is an index into the list
	//
	vector<char*>* files = *(vector<char*>**)lua_touserdata(L, lua_upvalueindex(1));
	int index = lua_tointeger(L, lua_upvalueindex(2));

	// if this is end of array (index >= length) then return 0 to indicate end
	//
	if(index >= (*files).length())
		return 0;

	// return the current filename
	//
	lua_pushstring(L, (*files)[index++]);

	// increment and store the index counter
	//
	lua_pushinteger(L, index);
	lua_replace(L, lua_upvalueindex(2));
	return 1;
}

// an error can happen at any point during the iteration, this is needed in
// order to use the correct C++ delete statement as well as store simple C++
// object data between the iterator and original function, as well as
// guarantee a cleanup of the object no matter what.
//
static int lua_DirectoryList_gc(lua_State* L)
{
	vector<char*>** ud = (vector<char*>**)lua_touserdata(L, 1);
	delete *ud;
	return 0;
}

int lua_DirectoryList(lua_State* L)
{
	luaL_checkstring(L, 1);
	const char* ext = luaL_optstring(L, 2, NULL);

	// clean the string parameter
	//
	lua_pushcfunction(L, lua_CleanString);
	lua_pushvalue(L, 1);
	lua_call(L, 1, 1);
	lua_replace(L, 1);

	vector<char*>** upval1;
	int curfile = 0; // second upvalue

	upval1 = (vector<char*>**)lua_newuserdata(L, sizeof(vector<char*>*));
	lua_newtable(L);
	lua_pushstring(L, "__gc");
	lua_pushcfunction(L, lua_DirectoryList_gc);
	lua_settable(L, -3);
	lua_setmetatable(L, -2);
	*upval1 = new vector<char*>();
	lua_pushinteger(L, curfile);

	bool ok = listdir(lua_tostring(L, 1), ext, **upval1);
	if(!ok)
	{
		lua_pop(L, 2); // remove the two upvalues
		return luaL_error(L, "Failed to retrieve directory listing from `%s'", lua_tostring(L, 1));
	}

	lua_pushcclosure(L, lua_DirectoryList_iter, 2);
	return 1;
}
ADDLUAUSERAPI(DirectoryList);

extern string homedir;

static int lua_GetHomeDirectory(lua_State* L)
{
    lua_pushstring(L, homedir);
    return 1;
}
ADDLUAPROTECTEDAPI(GetHomeDirectory);

static int lua_SetHomeDirectory(lua_State* L)
{
	luaL_checkstring(L, 1);
	copystring(homedir, lua_tostring(L, 1));
	return 0;
}
ADDLUAPROTECTEDAPI(SetHomeDirectory);
